@@ -1,7 +1,9 @@
-/* New simplified depmod without backwards compat stuff and not
-   requiring ksyms.
-
-   (C) 2002 Rusty Russell IBM Corporation
+/*
+ * New simplified depmod without backwards compat stuff and not
+ * requiring ksyms.
+ *
+ * (C) 2010 Jon Masters <jcm@jonmasters.org>, and others.
+ * (C) 2002 Rusty Russell IBM Corporation
  */
 #define _GNU_SOURCE /* asprintf */
 
@@ -59,6 +61,7 @@ struct module_search
 	size_t len;
 };
 
+static char sym_prefix;
 static unsigned int skipchars;
 static unsigned int make_map_files = 1; /* default to on */
 static unsigned int force_map_files = 0; /* default to on */
@@ -87,7 +90,12 @@ static inline unsigned int tdb_hash(const char *name)
 	return (1103515243 * value + 12345);
 }
 
-void add_symbol(const char *name, uint64_t ver, struct module *owner)
+static const char *skip_symprefix(const char *symname)
+{
+	return symname + (symname[0] == sym_prefix ? 1 : 0);
+}
+
+static void add_symbol(const char *name, uint64_t ver, struct module *owner)
 {
 	unsigned int hash;
 	struct symbol *new = NOFAIL(malloc(sizeof *new + strlen(name) + 1));
@@ -103,7 +111,7 @@ void add_symbol(const char *name, uint64_t ver, struct module *owner)
 
 static int print_unknown, check_symvers;
 
-struct module *find_symbol(const char *name, uint64_t ver,
+static struct module *find_symbol(const char *name, uint64_t ver,
 		const char *modname, int weak)
 {
 	struct symbol *s;
@@ -111,6 +119,7 @@ struct module *find_symbol(const char *name, uint64_t ver,
 	/* For our purposes, .foo matches foo.  PPC64 needs this. */
 	if (name[0] == '.')
 		name++;
+	name = skip_symprefix(name);
 
 	for (s = symbolhash[tdb_hash(name) % SYMBOL_HASH_SIZE]; s; s=s->next) {
 		if (streq(s->name, name))
@@ -129,7 +138,7 @@ struct module *find_symbol(const char *name, uint64_t ver,
 	return NULL;
 }
 
-void add_dep(struct module *mod, struct module *depends_on)
+static void add_dep(struct module *mod, struct module *depends_on)
 {
 	unsigned int i;
 
@@ -163,6 +172,7 @@ static void load_system_map(const char *filename)
 	/* eg. c0294200 R __ksymtab_devfs_alloc_devnum */
 	while (fgets(line, sizeof(line)-1, system_map)) {
 		char *ptr;
+		const char *cptr;
 
 		/* Snip \n */
 		ptr = strchr(line, '\n');
@@ -173,9 +183,12 @@ static void load_system_map(const char *filename)
 		if (!ptr || !(ptr = strchr(ptr + 1, ' ')))
 			continue;
 
+		/* Skip the space before symbol name */
+		cptr = skip_symprefix(ptr + 1);
+
 		/* Covers gpl-only and normal symbols. */
-		if (strstarts(ptr+1, ksymstr))
-			add_symbol(ptr+1+ksymstr_len, 0, NULL);
+		if (strstarts(cptr, ksymstr))
+			add_symbol(cptr + ksymstr_len, 0, NULL);
 	}
 
 	fclose(system_map);
@@ -202,14 +215,14 @@ static void load_module_symvers(const char *filename)
 			continue;
 
 		if (streq(where, "vmlinux"))
-			add_symbol(sym, strtoull(ver, NULL, 16), NULL);
+			add_symbol(skip_symprefix(sym), strtoull(ver, NULL, 16), NULL);
 	}
 
 	fclose(module_symvers);
 	add_fake_syms();
 }
 
-static struct option options[] = { { "all", 0, NULL, 'a' },
+static const struct option options[] = { { "all", 0, NULL, 'a' },
 				   { "quick", 0, NULL, 'A' },
 				   { "basedir", 1, NULL, 'b' },
 				   { "config", 1, NULL, 'C' },
@@ -222,6 +235,7 @@ static struct option options[] = { { "all", 0, NULL, 'a' },
 				   { "verbose", 0, NULL, 'v' },
 				   { "show", 0, NULL, 'n' },
 				   { "dry-run", 0, NULL, 'n' },
+				   { "symbol-prefix", 0, NULL, 'P' },
 				   { "help", 0, NULL, 'h' },
 				   { "version", 0, NULL, 'V' },
 				   { "warn", 0, NULL, 'w' },
@@ -233,7 +247,7 @@ static int is_version_number(const char *version)
 {
 	unsigned int dummy;
 
-	return (sscanf(version, "%u.%u.%u", &dummy, &dummy, &dummy) == 3);
+	return (sscanf(version, "%u.%u", &dummy, &dummy) == 2);
 }
 
 static int old_module_version(const char *version)
@@ -273,6 +287,7 @@ static void print_usage(const char *name)
 	"\t-e, --errsyms        Report not supplied symbols\n"
 	"\t-m, --map            Create the legacy map files\n"
 	"\t-n, --show           Write the dependency file on stdout only\n"
+	"\t-P, --symbol-prefix  Architecture symbol prefix\n"
 	"\t-V, --version        Print the release version\n"
 	"\t-v, --verbose        Enable verbose mode\n"
 	"\t-w, --warn		Warn on duplicates\n"
@@ -434,7 +449,7 @@ static const char *compress_path(const char *path, const char *basedir)
 	return path;
 }
 
-static void output_deps(struct module *modules,
+static int output_deps(struct module *modules,
 			FILE *out, char *dirname)
 {
 	struct module *i;
@@ -453,12 +468,13 @@ static void output_deps(struct module *modules,
 		}
 		fprintf(out, "\n");
 	}
+	return 1;
 }
 
 /* warn whenever duplicate module aliases, deps, or symbols are found. */
-int warn_dups = 0;
+static int warn_dups = 0;
 
-static void output_deps_bin(struct module *modules,
+static int output_deps_bin(struct module *modules,
 			FILE *out, char *dirname)
 {
 	struct module *i;
@@ -495,6 +511,8 @@ static void output_deps_bin(struct module *modules,
 	
 	index_write(index, out);
 	index_destroy(index);
+
+	return 1;
 }
 
 
@@ -743,7 +761,7 @@ static struct module *parse_modules(struct module *list)
 				check_symvers ? &symvers : NULL);
 		if (syms) {
 			for (j = 0; j < syms->cnt; j++)
-				add_symbol(syms->str[j],
+				add_symbol(skip_symprefix(syms->str[j]),
 					symvers ? symvers[j] : 0, i);
 			strtbl_free(syms);
 		}
@@ -769,7 +787,7 @@ static struct module *parse_modules(struct module *list)
 }
 
 /* Simply dump hash table. */
-static void output_symbols(struct module *unused, FILE *out, char *dirname)
+static int output_symbols(struct module *unused, FILE *out, char *dirname)
 {
 	unsigned int i;
 
@@ -786,9 +804,10 @@ static void output_symbols(struct module *unused, FILE *out, char *dirname)
 			}
 		}
 	}
+	return 1;
 }
 
-static void output_symbols_bin(struct module *unused, FILE *out, char *dirname)
+static int output_symbols_bin(struct module *unused, FILE *out, char *dirname)
 {
 	struct index_node *index;
 	unsigned int i;
@@ -817,9 +836,47 @@ static void output_symbols_bin(struct module *unused, FILE *out, char *dirname)
 	
 	index_write(index, out);
 	index_destroy(index);
+
+	return 1;
 }
 
-static void output_aliases(struct module *modules, FILE *out, char *dirname)
+static int output_builtin_bin(struct module *unused, FILE *out, char *dirname)
+{
+	struct index_node *index;
+	char *textfile, *line;
+	unsigned int linenum;
+	FILE *f;
+
+	nofail_asprintf(&textfile, "%s/modules.builtin", dirname);
+	if (!(f = fopen(textfile, "r"))) {
+		if (errno != ENOENT)
+			fatal("Could not open '%s': %s\n",
+					textfile, strerror(errno));
+		free(textfile);
+		return 0;
+	}
+	free(textfile);
+	index = index_create();
+
+	while ((line = getline_wrapped(f, &linenum)) != NULL) {
+		char *module = line;
+
+		if (!*line || *line == '#') {
+			free(line);
+			continue;
+		}
+		filename2modname(module, module);
+		index_insert(index, module, "", 0);
+		free(line);
+	}
+	fclose(f);
+	index_write(index, out);
+	index_destroy(index);
+
+	return 1;
+}
+
+static int output_aliases(struct module *modules, FILE *out, char *dirname)
 {
 	struct module *i;
 	struct elf_file *file;
@@ -834,13 +891,13 @@ static void output_aliases(struct module *modules, FILE *out, char *dirname)
 		filename2modname(modname, i->pathname);
 
 		/* Grab from old-style .modalias section. */
-		tbl = file->ops->load_strings(file, ".modalias", NULL, fatal);
+		tbl = file->ops->load_strings(file, ".modalias", NULL);
 		for (j = 0; tbl && j < tbl->cnt; j++)
 			fprintf(out, "alias %s %s\n", tbl->str[j], modname);
 		strtbl_free(tbl);
 
 		/* Grab from new-style .modinfo section. */
-		tbl = file->ops->load_strings(file, ".modinfo", NULL, fatal);
+		tbl = file->ops->load_strings(file, ".modinfo", NULL);
 		for (j = 0; tbl && j < tbl->cnt; j++) {
 			const char *p = tbl->str[j];
 			if (strstarts(p, "alias="))
@@ -849,9 +906,10 @@ static void output_aliases(struct module *modules, FILE *out, char *dirname)
 		}
 		strtbl_free(tbl);
 	}
+	return 1;
 }
 
-static void output_aliases_bin(struct module *modules, FILE *out, char *dirname)
+static int output_aliases_bin(struct module *modules, FILE *out, char *dirname)
 {
 	struct module *i;
 	struct elf_file *file;
@@ -870,7 +928,7 @@ static void output_aliases_bin(struct module *modules, FILE *out, char *dirname)
 		filename2modname(modname, i->pathname);
 
 		/* Grab from old-style .modalias section. */
-		tbl = file->ops->load_strings(file, ".modalias", NULL, fatal);
+		tbl = file->ops->load_strings(file, ".modalias", NULL);
 		for (j = 0; tbl && j < tbl->cnt; j++) {
 			alias = NOFAIL(strdup(tbl->str[j]));
 			underscores(alias);
@@ -883,7 +941,7 @@ static void output_aliases_bin(struct module *modules, FILE *out, char *dirname)
 		strtbl_free(tbl);
 
 		/* Grab from new-style .modinfo section. */
-		tbl = file->ops->load_strings(file, ".modinfo", NULL, fatal);
+		tbl = file->ops->load_strings(file, ".modinfo", NULL);
 		for (j = 0; tbl && j < tbl->cnt; j++) {
 			const char *p = tbl->str[j];
 			if (strstarts(p, "alias=")) {
@@ -901,15 +959,83 @@ static void output_aliases_bin(struct module *modules, FILE *out, char *dirname)
 	
 	index_write(index, out);
 	index_destroy(index);
+
+	return 1;
+}
+
+static int output_softdeps(struct module *modules, FILE *out, char *dirname)
+{
+	struct module *i;
+	struct elf_file *file;
+	struct string_table *tbl;
+	int j;
+
+	fprintf(out, "# Soft dependencies extracted from modules themselves.\n");
+	fprintf(out, "# Copy, with a .conf extension, to /etc/modprobe.d to use "
+		"it with modprobe.\n");
+	for (i = modules; i; i = i->next) {
+		char modname[strlen(i->pathname)+1];
+
+		file = i->file;
+		filename2modname(modname, i->pathname);
+
+		/* Grab from new-style .modinfo section. */
+		tbl = file->ops->load_strings(file, ".modinfo", NULL);
+		for (j = 0; tbl && j < tbl->cnt; j++) {
+			const char *p = tbl->str[j];
+			if (strstarts(p, "softdep="))
+				fprintf(out, "softdep %s %s\n",
+					modname, p + strlen("softdep="));
+		}
+		strtbl_free(tbl);
+	}
+	return 1;
+}
+
+static int output_devname(struct module *modules, FILE *out, char *dirname)
+{
+	struct module *m;
+
+	fprintf(out, "# Device nodes to trigger on-demand module loading.\n");
+	for (m = modules; m != NULL; m = m->next) {
+		struct string_table *tbl;
+		int i;
+		char type = '\0';
+		const char *devname = NULL;
+
+		tbl = m->file->ops->load_strings(m->file, ".modinfo", NULL);
+		for (i = 0; tbl && i < tbl->cnt; i++) {
+			const char *p = tbl->str[i];
+			unsigned int maj, min;
+
+			if (sscanf(p, "alias=char-major-%u-%u", &maj, &min) == 2)
+				type = 'c';
+			else if (sscanf(p, "alias=block-major-%u-%u", &maj, &min) == 2)
+				type = 'b';
+			else if (strstarts(p, "alias=devname:"))
+				devname = &p[strlen("alias=devname:")];
+
+			if (type && devname) {
+				char modname[strlen(m->pathname)+1];
+
+				filename2modname(modname, m->pathname);
+				fprintf(out, "%s %s %c%u:%u\n",
+					modname, devname, type, maj, min);
+				break;
+			}
+		}
+		strtbl_free(tbl);
+	}
+	return 1;
 }
 
 struct depfile {
-	char *name;
-	void (*func)(struct module *, FILE *, char *dirname);
+	const char *name;
+	int (*func)(struct module *, FILE *, char *dirname);
 	int map_file;
 };
 
-static struct depfile depfiles[] = {
+static const struct depfile depfiles[] = {
 	{ "modules.dep", output_deps, 0 }, /* This is what we check for '-A'. */
 	{ "modules.dep.bin", output_deps_bin, 0 },
 	{ "modules.pcimap", output_pci_table, 1 },
@@ -922,8 +1048,11 @@ static struct depfile depfiles[] = {
 	{ "modules.seriomap", output_serio_table, 1 },
 	{ "modules.alias", output_aliases, 0 },
 	{ "modules.alias.bin", output_aliases_bin, 0 },
+	{ "modules.softdep", output_softdeps, 0 },
 	{ "modules.symbols", output_symbols, 0 },
-	{ "modules.symbols.bin", output_symbols_bin, 0 }
+	{ "modules.symbols.bin", output_symbols_bin, 0 },
+	{ "modules.builtin.bin", output_builtin_bin, 0 },
+	{ "modules.devname", output_devname, 0 },
 };
 
 /* If we can't figure it out, it's safe to say "true". */
@@ -1074,8 +1203,7 @@ static int parse_config_file(const char *filename,
 			version = strsep_skipspace(&ptr, "\t ");
 			subdir = strsep_skipspace(&ptr, "\t ");
 
-			if (strcmp(version, kernelversion) != 0 &&
-			    strcmp(version, "*") != 0)
+			if (!regex_match(kernelversion, (const char *)version))
 				continue;
 
 			nofail_asprintf(&pathname, "%s%s%s/%s/%s.ko", basedir,
@@ -1221,9 +1349,9 @@ static void parse_toplevel_config(const char *filename,
 }
 
 /* Local to main, but not freed on exit.  Keep valgrind quiet. */
-struct module *list = NULL;
-struct module_search *search = NULL;
-struct module_overrides *overrides = NULL;
+static struct module *list = NULL;
+static struct module_search *search = NULL;
+static struct module_overrides *overrides = NULL;
 
 int main(int argc, char *argv[])
 {
@@ -1236,7 +1364,7 @@ int main(int argc, char *argv[])
 	if (native_endianness() == 0)
 		abort();
 
-	while ((opt = getopt_long(argc, argv, "aAb:C:E:F:euqrvnhVwm", options, NULL))
+	while ((opt = getopt_long(argc, argv, "aAb:C:E:F:euqrvnP:hVwm", options, NULL))
 	       != -1) {
 		switch (opt) {
 		case 'a':
@@ -1272,6 +1400,11 @@ int main(int argc, char *argv[])
 		case 'n':
 			doing_stdout = 1;
 			break;
+		case 'P':
+			if (optarg[1] != '\0')
+				fatal("-P only takes a single char\n");
+			sym_prefix = optarg[0];
+			break;
 		case 'h':
 			print_usage(argv[0]);
 			exit(0);
@@ -1296,7 +1429,7 @@ int main(int argc, char *argv[])
 	else if (system_map)
 		load_system_map(system_map);
 	else if (print_unknown) {
-		warn("-e needs -E or -F");
+		warn("-e needs -E or -F\n");
 		print_unknown = 0;
 	}
 
@@ -1366,9 +1499,10 @@ int main(int argc, char *argv[])
 	list = sort_modules(dirname,list);
 	list = parse_modules(list);
 
-	for (i = 0; i < sizeof(depfiles)/sizeof(depfiles[0]); i++) {
+	for (i = 0; i < ARRAY_SIZE(depfiles); i++) {
 		FILE *out;
-		struct depfile *d = &depfiles[i];
+		int res;
+		const struct depfile *d = &depfiles[i];
 		char depname[strlen(dirname) + 1 + strlen(d->name) + 1];
 		char tmpname[strlen(dirname) + 1 + strlen(d->name) +
 						strlen(".temp") + 1];
@@ -1388,12 +1522,18 @@ int main(int argc, char *argv[])
 			if (ends_in(depname, ".bin"))
 				continue;
 		}
-		d->func(list, out, dirname);
-		if (!doing_stdout) {
-			fclose(out);
+		res = d->func(list, out, dirname);
+		if (doing_stdout)
+			continue;
+		fclose(out);
+		if (res) {
 			if (rename(tmpname, depname) < 0)
 				fatal("Could not rename %s into %s: %s\n",
 					tmpname, depname, strerror(errno));
+		} else {
+			if (unlink(tmpname) < 0)
+				warn("Could not delete %s: %s\n",
+					tmpname, strerror(errno));
 		}
 	}
 
